class container:
    def __init__(self):
        self.all = None
        self.accepted = None
        self.rejected = None
        self.ϵ = None
        self.size = None
        self.n_accepted = None
